-- incomplete
--
-- EDITS:
-- - actually made the custom width params in font creation work
-- - added xsepwidth
-- - added func and predrawfunc to drawdata
-- - added forceuppercase
-- - added replaceTag to deal with rr tags
-- - added bold font rendering via <bold> with tags or \xC0
-- - fixed numerous issues
-- - fixed numerous warnings
-- - kill me
--
local WTXT2_VERSION = 1
local LOADED_VERSION = rawget(_G, "WTXT2")

-- Look, no need to load all this again when we don't need to.
if LOADED_VERSION != nil then
	if WTXT2_VERSION <= LOADED_VERSION then return end
end

local WTXT = {}

local function devprint(...)
	if not devparm then return end
	local argcount = select('#', ...)
	local argtbl = {...}
	local bigstr = ""
	for argix = 1,argcount do
		local arg = argtbl[argix]
		bigstr = $ .. tostring(arg) .. (argix ~= argcount and "\t" or "")
	end
	print("\139[WTXT]\128\t" .. bigstr)
end
local function devprint_unpaused(...)
	if paused then return end
	devprint(...)
end

local fonts = {}

local function buildFont(v, font)
	devprint("buildFont called...", font.name)
	-- (re)builds a font and its glyphs
	-- should normally be done only once, on font creation
	-- assumes font is good!!! issues should be handled by createNewFont
	
	local glyphs = {}
	
	local glyphExistCount = 0
	local maxWidth = 0
	local maxHeight = 0
	
	for ascii = 0,255 do -- for each possible glyph in a font
		local glyph = {}
		
		local patchName = string.format("%s%03d", font.prefix, ascii)
		glyph.exists = v.patchExists(patchName)
		
		if glyph.exists then
			glyphExistCount = $+1
			local patch = v.cachePatch(patchName)
			glyph.patch = patch
			glyph.width = patch.width
			glyph.height = patch.height
			
			maxHeight = max($, patch.height)
			maxWidth = max($, patch.width)
		end
		
		-- no more editing past this point
		--glyphs[ascii] = setmetatable(glyph, _frozen)
		glyphs[ascii] = glyph
	end
	
	--local avgWidth =
	
	-- some other font work
	-- we can afford to be slow and precise. this is run once
	font.validGlyphs = glyphExistCount
	font.maxWidth = maxWidth
	font.maxHeight = maxHeight
	
	if font.sepwidth == nil then
		devprint("sepwidth missing, set to 0")
		font.sepwidth = 0
	end
	
	if font.xsepwidth == nil then
		devprint("xsepwidth missing, set to 0")
		font.xsepwidth = 0
	end
	
	if font.monowidth == nil then
		devprint("monowidth missing, using widest glyph's width", maxWidth)
		font.monowidth = maxWidth
	end
	
	if font.spacewidth == nil then
		local half = font.monowidth / 2
		devprint("spacewidth missing, using half of monowidth", font.monowidth, half)
		font.spacewidth = half
	end
	
	if font.ysep == nil then
		devprint("ysep missing, set to 0")
		font.ysep = 0
	end
	
	font.glyphs = glyphs
	font.sane = true -- we just built this, so it should be good!
	
	devprint("buildFont success!")
	return true
end

local function fontSentinel(v)
	-- when called:
	-- - check font table for new additions and parse them
	-- - see if we have to regenerate any fonts for some reason
	
	for fontix = 1,#fonts do
		local font = fonts[fontix]
		if not font.sane then -- uh oh!
			devprint("insane font found. delegate > buildFont()")
			buildFont(v,font)
		end
	end
	
	return true
end

local function createNewFont(info)
	devprint("createNewFont called...")
	local font = {}
	
	font.info = info
	
	if type(info.prefix) ~= "string" then error("prefix must be a string",2) end
	if info.prefix:len() ~= 5 then error("prefix length must be exactly 5 characters",2) end
	font.prefix = info.prefix
	
	-- amperbee forgot
	font.sepwidth = info.sepwidth
	font.xsepwidth = info.xsepwidth
	font.monowidth = info.monowidth
	font.spacewidth = info.spacewidth
	-- function(scale, patchwidth)
	font.dimfunc = info.dimfunc
	font.ysep = info.ysep
	font.forceUppercase = info.forceUppercase or false
	font.extended = info.extended
	font.fallbackColors = info.fallbackColors
	
	-- TODO maybe make some more checks to make sure font is somewhat good
	
	font.sane = false -- this will get fontSentinel to build the font
	fonts[#fonts+1] = font
	
	rawset(_G, "FONT_"..info.name, #fonts)
	devprint("createNewFont success!")
	
	return true
end


--##################
--## DRAWING TEXT ##
--##################

local flagval = 0; local function flagup() flagval = $+1; return 1<<(flagval-1) end
rawset(_G, "DF_INTPOS", flagup())
	-- indicates you're not using fixed point coords
	-- multiply coords by FRACUNIT
	
rawset(_G, "DF_MONOSPACE", flagup())
	-- if set, uses monospace
	
rawset(_G, "DF_ADVANCED", flagup())
	-- if set, enables more computationally expensive features
	-- TODO what?
	-- most useful for short strings
	
--rawset(_G, "DF_WHITESPACEGLYPH", flagup())
	-- if set, when parsing whitespace, place a fake glyph with no associated patch
	-- useless for normal use, but if you do something that counts characters drawn,
	-- might be useful
	
rawset(_G, "DF_WORDWRAP", flagup())
	-- if wrapping is enabled, this flag changes wrapping behavior
	-- by attempting to avoid breaking words.
	
rawset(_G, "DF_STRICT", flagup())
	-- if the drawer attempts to use a glyph that does not have a graphical representation,
	-- and it is not a character with special interpretation (\n, space, ???),
	-- raise error

--rawset(_G, "DF_SPECIALGLYPHS", flagup())
	-- if enabled, special glyphs (space, newline...) will draw their respective glyph.
	-- caveats:
	-- - if there is a space glyph, spacewidth is ignored
	-- - a newline glyph will be drawn, and line will shift afterwards
	-- nah on second thought this is stupid and useless, just use the other 255 chars

local function prepareLine()
	return {
		x = 0,
		y = 0,
		width = 0,
		height = 0,
		lastWordStart = 0, -- ix of last word char preceded by whitespace. 0 if this has not happened yet
	}
end
local function concludeLine(board, newline)
	local line = board.workline
	board[#board+1] = line
	-- TODO: calculate line width and offset, mayhaps
	
	--line.x = FixedMul(line.curwidth, FU - board.halign)
	--line.x = line.curwidth
	line.x = 0 -- for now
	
	local lastPiece = line[#line]
	-- no need to calculate the width if theres nothing here
	if lastPiece ~= nil then
		line.width = lastPiece.x + lastPiece.width -- sensible
	else
		line.width = 0
		line.height = board.backupHeight
	end
	
	--board.widestLineWidth = max(line.width, $)
	board.autowidth = max(line.width, $)
	board.autoheight = $ + line.height
	
	if newline then
		-- line completed, perform final adjustments
		local nextline = prepareLine()
		nextline.y = line.y + line.height -- move this new line downwards a bit, ya?
		board.workline = nextline
		return nextline
	else board.workline = nil end
end
local function makePiece(ascii)
end
local cmapfuckery = {
	-- colors
	[0x80] = SKINCOLOR_NONE,
	[0x81] = SKINCOLOR_PURPLE,
	[0x82] = SKINCOLOR_YELLOW,
	[0x83] = SKINCOLOR_GREEN,
	[0x84] = SKINCOLOR_BLUE,
	[0x85] = SKINCOLOR_RED,
	[0x86] = SKINCOLOR_GREY,
	[0x87] = SKINCOLOR_ORANGE,
	[0x88] = SKINCOLOR_JAWZ,
	[0x89] = SKINCOLOR_LAVENDER,
	[0x8A] = SKINCOLOR_GOLD,
	[0x8B] = SKINCOLOR_AQUAMARINE,
	[0x8C] = SKINCOLOR_MAGENTA,
	[0x8D] = SKINCOLOR_PINK,
	[0x8E] = SKINCOLOR_BROWN,
	[0x8F] = SKINCOLOR_SEPIA,
}
-- Does not take into account sepwidth for now
local function wordWrap(str, f, w, hscale, dflags, wflags)
	local font = fonts[f] -- TODO check this exists before using
	
	-- Full uppercase if told to do so
	if font.forceUppercase or (dflags & V_FORCEUPPERCASE) then
		str = str:upper()
	end
	
	local crsr = 0 -- current character from the string being processed. 1 is first char
	local lastcrsr = 0 -- last cursor
	local crsrMax = #str -- when to stop
	
	local monospaceMode = wflags & DF_MONOSPACE > 0 -- advanced mode (cool stuff)
	
	local cx = 0
	local lastWord = 0
	local splitcount = 0
	local tooManySplits = 8
	
	while crsr < crsrMax do
		crsr = $+1; local chr = str:byte(crsr,crsr) -- advance cursor and extract a char as byte
		
		local pw = 0
		
		if chr == 0x0A then -- reset shit
			cx = 0
			lastWord = 0
		end
		if chr < 32 or chr > 127 then
			-- do nothing
		elseif chr == 0x20 then -- " ", space
			pw = font.spacewidth
			lastWord = crsr
		elseif font.glyphs[chr] and font.glyphs[chr].exists then
			local glyph = font.glyphs[chr]
			pw = FixedMul((glyph.width+font.xsepwidth) * FRACUNIT, hscale)/FRACUNIT
			if font.dimfunc then
				pw = font.dimfunc(hscale, (glyph.width+font.xsepwidth) * FRACUNIT)/FRACUNIT
			end
		end
		
		if monospaceMode and chr >= 32 and chr <= 127 then
			pw = FixedMul((font.monowidth) * FRACUNIT, hscale)/FRACUNIT
			--piece.height = (glyph.height) * FRACUNIT -- :shrug:
		end
		
		-- I dont get why I have to halve this but oh well
		cx = $ + pw/2
		
		if w then -- if board has a width limit. see if we can fit piece first
			if cx > w and splitcount < tooManySplits then
				--print(string.format("split!: cx: %d, w: %d", cx, w))
				crsr = lastWord
				local str1 = string.sub(str, 1, crsr)
				local str2 = string.sub(str, crsr+1, #str)
				str = str1.."\n"..str2
				--crsr = lastWord+1
				crsrMax = #str
				cx = 0
				splitcount = $+1
			end
		end
	end
	
	return str
end

local function makeBoard(drawdata)
	-- takes a table with text and formatting detail
	-- outputs a "board" of pieces (or patches) to be drawn
	
	local board = {
		x = 0,
		y = 0,
		width = drawdata.width or -1,
		height = -1,
		autowidth = 0,
		autoheight = 0,
	}
	local str = drawdata.text
	local font = fonts[drawdata.font] -- TODO check this exists before using
	local boldFont = drawdata.boldFont and fonts[drawdata.boldFont]
	local useFont = font
	
	-- Full uppercase if told to do so
	if font.forceUppercase or (drawdata.dflags & V_FORCEUPPERCASE) then
		str = string.upper($)
	end
	
	local crsr = 0 -- current character from the string being processed. 1 is first char
	local lastcrsr = 0 -- last cursor
	local crsrMax = #str -- when to stop
	
	local strict = drawdata.wflags & DF_STRICT > 0 -- strict mode (missing glyphs raise error)
	local advanced = drawdata.wflags & DF_ADVANCED > 0 -- advanced mode (cool stuff)
	
	local monospaceMode = drawdata.wflags & DF_MONOSPACE > 0 -- advanced mode (cool stuff)
	local wordWrap = drawdata.wflags & DF_WORDWRAP > 0 -- word wrapping is enabled
	local widthLimit = drawdata.width ~= nil -- glyphs should be bound within box
	
	local lastTextColor = 0 -- last text color, for use in colormaps
	local fallbackTextColor = 0
	local boldFontToggle = false -- toggle fonts
	
	--local curchr = 1 -- 
	--local curheight = 0 --
	
	board.workline = prepareLine() -- current line being worked on
	board.backupHeight = font.monowidth * FRACUNIT -- TODO: Separate font attribute???
	
	-- begin parsing every character into a drawable patch
	-- TODO should keep track of last word character for DF_WORDWRAP to be able to work
	while crsr < crsrMax do
		crsr = $+1; local chr = str:byte(crsr,crsr) -- advance cursor and extract a char as byte
		local line = board.workline
		
		local notFirstSpace = (#line > 0 and font.sepwidth*FRACUNIT or 0)
		-- add a little extra space if this is not the first character and sepwidth is nonzero
		
		local piece = {
			-- these positions should be relative to line start pos
			x = line.width + notFirstSpace, -- start at the right side of the last piece + separator
			y = 0, -- should always be zero :shrug:
			width = 0, -- "sensible" default
			height = 0, -- same here
			hscale = drawdata.hscale,
			vscale = drawdata.vscale,
			xoff = 0, -- x offset
			yoff = 0, -- ditto
			num = crsr, -- current number for safekeeping
			color = lastTextColor,
			textColor = fallbackTextColor,
			useFallback = useFont.fallbackColors ~= nil,
			colormap = nil,
			bold = boldFontToggle,
		}
		
		local finishLine = false
		-- if this char is special, parse special properties first
		-- TODO tabulation
		if chr == 0x20 then -- " ", space
			-- special: just add distance, do not draw. drop piece
			piece.whitespace = true
			piece.width = FixedMul(font.spacewidth * FRACUNIT, piece.hscale) -- piece gets dropped anyways, might change later
			
			-- TODO unless it has an associated patch OR specific flag is set.
			
		elseif chr == 0x0A then -- "\n", line feed
			piece.whitespace = true
			piece.height = board.backupHeight
			finishLine = true
		
		elseif (chr & 0xF0) == 0x80 then -- soc color
			piece.whitespace = true
			lastTextColor = cmapfuckery[chr]
			fallbackTextColor = (chr & 0x7f)
		
		elseif (chr == 0xC0) then -- toggle bold font if available
			piece.whitespace = true
			boldFontToggle = not boldFontToggle
			if boldFontToggle and boldFont then
				useFont = boldFont
			else
				useFont = font
			end
			
		else -- everyday piece
			local glyph = useFont.glyphs[chr] -- a glyph from the font
			if glyph == nil then -- piece does not exist.
				piece.whitespace = true
			else
				piece.glyph = glyph
				if glyph.exists then
					piece.width = FixedMul((glyph.width+useFont.xsepwidth) * FRACUNIT, piece.hscale)
					piece.height = FixedMul((glyph.height+useFont.ysep) * FRACUNIT, piece.vscale)
				end
			end
		end
		
		if useFont.dimfunc and piece.glyph and (not piece.whitespace) then
			piece.width = useFont.dimfunc(piece.hscale, (piece.glyph.width+useFont.xsepwidth) * FRACUNIT)
		end
		
		if monospaceMode then
			piece.width = FixedMul((useFont.monowidth) * FRACUNIT, piece.hscale)
			--piece.height = (glyph.height) * FRACUNIT -- :shrug:
		end
		
		local totalwidth = notFirstSpace + piece.width
		
		if not piece.whitespace then
			if widthLimit then -- if board has a width limit. see if we can fit piece first
				if line.width + totalwidth > board.width then
					-- this character would go past bounds. que for next line...
					if wordWrap then
						-- TODO solution is ugly
						-- look for an alternative
						-- commit what i have
					else
						line = concludeLine(board, true)
						piece.x = 0 -- congrats on the promotion, buddy!
						notFirstSpace = 0 -- null space bonus, we started a new line
					end
				end
			end
			
			line[#line+1] = piece -- place piece in line
		else
			line.lastWordStart = #line -- get marked buddy
		end
		
		line.width = $ + notFirstSpace + piece.width -- this line has become wider
		line.height = max($, piece.height) -- lol.
		
		-- Alright, we're done here.
		lastcrsr = $+1
		
		if finishLine then concludeLine(board, true) end
	end -- no more characters
	
	-- if there was a line in progress, complete it
	concludeLine(board, false)
	
	-- set board bounds if not set by drawdata
	if drawdata.width == nil then board.width = board.autowidth end
	if drawdata.height == nil then board.height = board.autoheight end
	
	-- align lines now
	-- TODO: if boards have variable height, consider that. for now, ignore
	local effhalign = (drawdata.halign/2) + FRACUNIT/2 -- f(x) = 0.5x + 0.5
	local effvalign = (drawdata.valign/2) + FRACUNIT/2
	
	-- align lines horizontally, then again for board
	for lineix = 1,#board do
		local line = board[lineix]
		local halignpad = FixedMul(board.width - line.width, effhalign)
		--local valignpad = FixedMul(board.width - line.height, effvalign)
		line.x = $ + halignpad
	end
	board.x = FixedMul(-board.width, effhalign)
	board.y = FixedMul(-board.height, effvalign)
	
	return board -- drawBoard would use this
end

local function co_boardIterator(board)
	
	for lineix = 1,#board do
		local line = board[lineix]
		for crsr = 1,#line do
			local piece = line[crsr]
			
			coroutine.yield(piece, line)
		end
	end
	
end

local function drawBoard(v, board, drawdata)
	-- takes a "board" of patches that can be drawn
	-- and draws them
	
	-- If you wanna draw something before the text draw it here
	if drawdata.func then drawdata.func(drawdata, board) end
	
	local corou = coroutine.create(co_boardIterator)
	for piece,line in function()
		-- iterate pieces, returning each piece and respective line
		local code,piece,line = coroutine.resume(corou, board)
		return piece,line
	end do
		-- :thinking:
		if drawdata.predrawfunc then drawdata.predrawfunc(drawdata, board, line, piece) end
		
		local finalx,finaly =
			drawdata.x + board.x + line.x + piece.x + piece.xoff,
			drawdata.y + board.y + line.y + piece.y + piece.yoff
		;
		-- jandle soc colour
		local cmap = drawdata.colormap
		if piece.color and not piece.useFallback then
			cmap = v.getColormap(TC_RAINBOW, piece.color)
		elseif piece.useFallback then
			cmap = v.getStringColormap(piece.textColor << V_CHARCOLORSHIFT)
		end
		if piece.colormap then
			cmap = piece.colormap
		end
		if piece.glyph.patch then
			v.drawStretched(
				finalx, finaly,
				piece.hscale, piece.vscale,
				piece.glyph.patch,
				drawdata.dflags, cmap -- TODO: piece flags and piece colormap?
			)
		end
	end
	
end

local btnTags = {
	["up"] = 0x00,
	["down"] = 0x01,
	["right"] = 0x02,
	["left"] = 0x03,
	
	["dpad"] = 0x04,
	
	["r"] = 0x07,
	["l"] = 0x08,
	["start"] = 0x09,
	
	["a"] = 0x0A,
	["b"] = 0x0B,
	["c"] = 0x0C,
	
	["x"] = 0x0D,
	["y"] = 0x0E,
	["z"] = 0x0F,
}

local v_drawTags = {
	-- colors
	["white"] = 0x80,
	["purple"] = 0x81,
	["yellow"] = 0x82,
	["green"] = 0x83,
	["blue"] = 0x84,
	["red"] = 0x85,
	["gray"] = 0x86,
	["orange"] = 0x87,
	["sky"] = 0x88,
	["lavender"] = 0x89,
	["gold"] = 0x8A,
	["aqua"] = 0x8B,
	["magenta"] = 0x8C,
	["pink"] = 0x8D,
	["brown"] = 0x8E,
	["tan"] = 0x8F,
	-- special tags
	["bold"] = 0xC0,
	["wait"] = 0xC1,
}

-- Add button tags lmao
do
	for k, v in ipairs(btnTags) do
		table.insert(v_drawTags, k, 0xB0 | v)
		table.insert(v_drawTags, k.."_animated", 0xA0 | v)
		table.insert(v_drawTags, k.."_pressed", 0x90 | v)
	end
end

-- Find tags oooo
local function replaceTags(text)
	local lasttag = ""
	for tag in string.gmatch(text, "%<.-%>") do
		-- Find the start of this tag to parse it
		local s, p = string.find(text, tag)
		-- Extract the tag
		local t = string.sub(tag, 2, string.len(tag) - 1)
		-- Replace it with the right control code
		if v_drawTags[t] then
			local repl = v_drawTags[t]
			-- Mimic RR v_draw color tag behavior
			if lasttag == tag and repl < 0x90 then
				repl = v_drawTags["white"]
			end
			text = text:gsub(tag, string.char(repl))
		end
		lasttag = tag
	end
	return text
end

local function writeText(v, drawdata)
	fontSentinel(v) -- ensure fonts are ready to go
	
	-- should probably do some checks here to make sure drawdata is good
	drawdata.wflags = $ or 0
	if drawdata.scale == nil then drawdata.scale = FRACUNIT end
	if drawdata.hscale == nil then drawdata.hscale = drawdata.scale end
	if drawdata.vscale == nil then drawdata.vscale = drawdata.scale end
	
	if drawdata.halign == nil then drawdata.halign = -FRACUNIT end
	if drawdata.valign == nil then drawdata.valign = -FRACUNIT end
	
	drawdata.dflags = $ or 0
	
	local intpos = drawdata.wflags & DF_INTPOS > 0 -- advanced mode (cool stuff)
	if intpos then
		drawdata.x = $ * FRACUNIT
		drawdata.y = $ * FRACUNIT
	end
	
	local board = makeBoard(drawdata) -- prepare the board
	drawBoard(v, board, drawdata) -- draw it to the screen
end

WTXT.writeText = writeText
WTXT.createNewFont = createNewFont
WTXT.fontSentinel = fontSentinel
-- Insert existing fonts
if WTXT.fonts then
	for k, v in pairs(WTXT.fonts) do
		table.insert(fonts, v)
	end
end
WTXT.fonts = fonts
WTXT.replaceTags = replaceTags
WTXT.wordWrap = wordWrap

rawset(_G, "WTXT", WTXT)
rawset(_G, "WTXT2", WTXT2_VERSION)

devprint("I'm ready!")









